@@ -22,6 +22,7 @@ from jaxopt._src.anderson import AndersonAcceleration
 from jaxopt._src.anderson_wrapper import AndersonWrapper
 from jaxopt._src.armijo_sgd import ArmijoSGD
 from jaxopt._src.backtracking_linesearch import BacktrackingLineSearch
+from jaxopt._src.bfgs import BFGS
 from jaxopt._src.bisection import Bisection
 from jaxopt._src.block_cd import BlockCoordinateDescent
 from jaxopt._src.cd_qp import BoxCDQP
@@ -41,7 +42,6 @@ from jaxopt._src.osqp import OSQP
 from jaxopt._src.polyak_sgd import PolyakSGD
 from jaxopt._src.projected_gradient import ProjectedGradient
 from jaxopt._src.proximal_gradient import ProximalGradient
-from jaxopt._src.quadratic_prog import QuadraticProgramming
 from jaxopt._src.scipy_wrappers import ScipyBoundedLeastSquares
 from jaxopt._src.scipy_wrappers import ScipyBoundedMinimize
 from jaxopt._src.scipy_wrappers import ScipyLeastSquares
