@@ -37,6 +37,7 @@ from jaxopt.tree_util import tree_scalar_mul
 from jaxopt.tree_util import tree_sub
 from jaxopt.tree_util import tree_sum
 from jaxopt.tree_util import tree_l2_norm
+from jaxopt._src.tree_util import tree_single_dtype
 
 
 def inv_hessian_product_leaf(v: jnp.ndarray,
@@ -123,7 +124,7 @@ def compute_gamma(s_history: Any, y_history: Any, last: int):
 
 
 def init_history(pytree, history_size):
-  fun = lambda leaf: jnp.zeros((history_size,) + leaf.shape)
+  fun = lambda leaf: jnp.zeros((history_size,) + leaf.shape, dtype=leaf.dtype)
   return tree_map(fun, pytree)
 
 
@@ -141,7 +142,9 @@ class LbfgsState(NamedTuple):
   s_history: Any
   y_history: Any
   rho_history: jnp.ndarray
+  gamma: jnp.ndarray
   aux: Optional[Any] = None
+  failed_linesearch: bool = False
 
 
 @dataclass(eq=False)
@@ -150,8 +153,19 @@ class LBFGS(base.IterativeSolver):
 
   Attributes:
     fun: a smooth function of the form ``fun(x, *args, **kwargs)``.
-    has_aux: whether function fun outputs one (False) or more values (True).
-      When True it will be assumed by default that fun(...)[0] is the objective.
+    value_and_grad: whether ``fun`` just returns the value (False) or both
+      the value and gradient (True).
+    has_aux: whether ``fun`` outputs auxiliary data or not.
+      If ``has_aux`` is False, ``fun`` is expected to be
+        scalar-valued.
+      If ``has_aux`` is True, then we have one of the following
+        two cases.
+      If ``value_and_grad`` is False, the output should be
+      ``value, aux = fun(...)``.
+      If ``value_and_grad == True``, the output should be
+      ``(value, aux), grad = fun(...)``.
+      At each iteration of the algorithm, the auxiliary outputs are stored
+        in ``state.aux``.
 
     maxiter: maximum number of proximal gradient descent iterations.
     tol: tolerance of the stopping criterion.
@@ -160,6 +174,8 @@ class LBFGS(base.IterativeSolver):
       or a callable specifying the **positive** stepsize to use at each iteration.
     linesearch: the type of line search to use: "backtracking" for backtracking
       line search or "zoom" for zoom line search.
+    stop_if_linesearch_fails: whether to stop iterations if the line search fails.
+      When True, this matches the behavior of core JAX.
     maxls: maximum number of iterations to use in the line search.
     decrease_factor: factor by which to decrease the stepsize during line search
       (default: 0.8).
@@ -191,6 +207,7 @@ class LBFGS(base.IterativeSolver):
   """
 
   fun: Callable
+  value_and_grad: bool = False
   has_aux: bool = False
 
   maxiter: int = 500
@@ -198,6 +215,7 @@ class LBFGS(base.IterativeSolver):
 
   stepsize: Union[float, Callable] = 0.0
   linesearch: str = "zoom"
+  stop_if_linesearch_fails: bool = False
   condition: str = "strong-wolfe"
   maxls: int = 15
   decrease_factor: float = 0.8
@@ -217,6 +235,14 @@ class LBFGS(base.IterativeSolver):
 
   verbose: bool = False
 
+  def _cond_fun(self, inputs):
+    _, state = inputs[0]
+    if self.verbose:
+      print("error:", state.error)
+    # We continue the optimization loop while the error tolerance is not met and,
+    # either failed linesearch is disallowed or linesearch hasn't failed.
+    return (state.error > self.tol) & (~self.stop_if_linesearch_fails | ~state.failed_linesearch)
+
   def init_state(self,
                  init_params: Any,
                  *args,
@@ -234,15 +260,17 @@ class LBFGS(base.IterativeSolver):
       _, aux = self.fun(init_params, *args, **kwargs)
     else:
       aux = None
-
+    dtype = tree_single_dtype(init_params)
     return LbfgsState(iter_num=jnp.asarray(0),
                       value=jnp.asarray(jnp.inf),
-                      stepsize=jnp.asarray(self.max_stepsize),
+                      stepsize=jnp.asarray(self.max_stepsize, dtype=dtype),
                       error=jnp.asarray(jnp.inf),
                       s_history=init_history(init_params, self.history_size),
                       y_history=init_history(init_params, self.history_size),
-                      rho_history=jnp.zeros(self.history_size),
-                      aux=aux)
+                      rho_history=jnp.zeros(self.history_size, dtype=dtype),
+                      gamma=jnp.asarray(1.0, dtype=dtype),
+                      aux=aux,
+                      failed_linesearch=jnp.asarray(False))
 
   def update(self,
              params: Any,
@@ -262,20 +290,15 @@ class LBFGS(base.IterativeSolver):
     (value, aux), grad = self._value_and_grad_with_aux(params, *args, **kwargs)
 
     start = state.iter_num % self.history_size
-    previous = (start + self.history_size - 1) % self.history_size
-
-    if self.use_gamma:
-      gamma = compute_gamma(state.s_history, state.y_history, previous)
-    else:
-      gamma = 1.0
 
     product = inv_hessian_product(pytree=grad, s_history=state.s_history,
                                   y_history=state.y_history,
-                                  rho_history=state.rho_history, gamma=gamma,
+                                  rho_history=state.rho_history, gamma=state.gamma,
                                   start=start)
     descent_direction = tree_scalar_mul(-1.0, product)
 
-    if not isinstance(self.stepsize, Callable) and self.stepsize <= 0:
+    use_linesearch = not isinstance(self.stepsize, Callable) and self.stepsize <= 0
+    if use_linesearch:
       # with line search
 
       if self.linesearch == "backtracking":
@@ -311,7 +334,8 @@ class LBFGS(base.IterativeSolver):
         new_grad = ls_state.g_k
         # FIXME: zoom_linesearch currently doesn't return new_params
         # so we have to recompute it.
-        new_params = tree_add_scalar_mul(params, new_stepsize, descent_direction)
+        t = new_stepsize.astype(tree_single_dtype(params))
+        new_params = tree_add_scalar_mul(params, t, descent_direction)
 
       else:
         raise ValueError("Invalid name in 'linesearch' option.")
@@ -337,6 +361,16 @@ class LBFGS(base.IterativeSolver):
     y_history = update_history(state.y_history, y, last)
     rho_history = update_history(state.rho_history, rho, last)
 
+    if self.use_gamma:
+      gamma = compute_gamma(s_history, y_history, last)
+    else:
+      gamma = jnp.array(1.0)
+
+    if use_linesearch and self.linesearch == "zoom":
+      failed_linesearch = ls_state.failed
+    else:  # backtracking linesearch doesn't support failed state yet
+      failed_linesearch = jnp.asarray(False)
+
     new_state = LbfgsState(iter_num=state.iter_num + 1,
                            value=new_value,
                            stepsize=jnp.asarray(new_stepsize),
@@ -344,33 +378,27 @@ class LBFGS(base.IterativeSolver):
                            s_history=s_history,
                            y_history=y_history,
                            rho_history=rho_history,
+                           gamma=gamma,
                            # FIXME: we should return new_aux here but
                            # BacktrackingLineSearch currently doesn't support
                            # an aux.
-                           aux=aux)
+                           aux=aux,
+                           failed_linesearch=failed_linesearch)
 
     return base.OptStep(params=new_params, state=new_state)
 
   def optimality_fun(self, params, *args, **kwargs):
     """Optimality function mapping compatible with ``@custom_root``."""
-    return self._grad_fun(params, *args, **kwargs)
+    return self._value_and_grad_fun(params, *args, **kwargs)[1]
 
   def _value_and_grad_fun(self, params, *args, **kwargs):
     (value, aux), grad = self._value_and_grad_with_aux(params, *args, **kwargs)
     return value, grad
 
-  def _grad_fun(self, params, *args, **kwargs):
-    return self._value_and_grad_fun(params, *args, **kwargs)[1]
-
   def __post_init__(self):
-    if self.has_aux:
-      self._fun = lambda *a, **kw: self.fun(*a, **kw)[0]
-      fun_with_aux = self.fun
-    else:
-      self._fun = self.fun
-      fun_with_aux = lambda *a, **kw: (self.fun(*a, **kw), None)
-
-    self._value_and_grad_with_aux = jax.value_and_grad(fun_with_aux,
-                                                       has_aux=True)
+    _, _, self._value_and_grad_with_aux = \
+      base._make_funs_with_aux(fun=self.fun,
+                               value_and_grad=self.value_and_grad,
+                               has_aux=self.has_aux)
 
     self.reference_signature = self.fun
